/*
   intersection.c: this file contains functions:
   checkIntersections : checks if two triangles intersect in space
   DoesIntersect : checks if a segment intersects a triangle.
   other functions 

   It is written by Joseph O'Rourke (Professor of Computer Science,
   Smith College) and Carole Gitlin (BA, Computer Science, Smith College), 
   with a few additions written by Gill Barequet (Israel) and Anna 
   Lysyanskaya (Computer Science, Smith College).
*/


#include <stdio.h>
#include <math.h>

#define MAXNUM 20000
#define MIN(A,B) ((A < B) ? A : B)
#define TRUE 1
#define FALSE 0
#define BAD 0
#define GOOD 1
#define boolean int


typedef int pVertex[3];
typedef pVertex Triangle[3];


int DoesIntersect (Triangle T, pVertex, pVertex, boolean sharedsides);
double VolumeD(pVertex a, pVertex b, pVertex c, pVertex p);
int OppositeSides (Triangle t1, pVertex v0, pVertex v1);
int Volume6 (pVertex, pVertex, pVertex, pVertex);
void subtract (pVertex X, pVertex a, pVertex b);
void printVertex (pVertex a, char *s);
int coplanarCheck(Triangle T, pVertex a, pVertex b, boolean sharedsides);
int surrounded (pVertex a, pVertex b, Triangle T, 
		double intersections[][3], int x, int y);
int inside (pVertex a, pVertex b, double intersection[3], int
	    sharedsides, int x, int y);
int insideF (double a[3], double b[3], pVertex intersection, int
	    sharedsides, int x, int y);
void	Cross( pVertex a, pVertex b, pVertex c);
int Dot( pVertex a, pVertex b );
void initIntersections (double intersections[3][3]);

/***************************************************************************** 
   checkIntersection: 

   written by Gill Barequet, modified by Anna Lysyanskaya

   inputs: two triangles, and boolean sharedsides. If sharedsides is 
           TRUE, that means that triangles may share sides. Otherwise
	   they may not.

******************************************************************************/

int checkIntersection (Triangle T1, Triangle T2, boolean sharesides) {
 

   int i, j;

   if (sharesides) {
     /* We have to check if the triangles are identical separately: */
     for (i = 0; i< 3; i++) {
       for (j = 0; j < 3; j++) {
	 if (T1[i][j] != T2[i][j]) goto check;
       }
     }
     return TRUE;
   }
   
 check:   for (i = 0; i < 3; i ++)
   if (DoesIntersect (T1, T2 [i], T2 [(i + 1) % 3], sharesides)) {
     return (TRUE);
   }
   for (i = 0; i < 3; i ++)
     if (DoesIntersect (T2, T1 [i], T1 [(i + 1) % 3], sharesides)) {
       return (TRUE);
     }
   return (FALSE);
 }               





/*
	Written by Joseph O'Rourke and Carole Gitlin.
*/

/*
	If the segment lies strictly on one side of the plane determined 
	by the triangle, they do not intersect.
*/
int OppositeSides(Triangle t1, pVertex v0, pVertex v1)
{
  int vol1, vol2;
  double vol1d, vol2d;
  
  vol1 = Volume6(t1[0],t1[1],t1[2],v0);
  vol1d= VolumeD(t1[0],t1[1],t1[2],v0);
  if (fabs(vol1 - vol1d) >= 1.0)
    {
      printf("Possible Integer overflow in OppositeSides.\n");
      printf("volume1 = %d     volume1d = %f\n",vol1,vol1d);
      exit(1);
    }
  vol2 = Volume6(t1[0],t1[1],t1[2],v1);
  vol2d= VolumeD(t1[0],t1[1],t1[2],v1);
  if (fabs (vol2 - vol2d) >= 1.0)
    {
      printf("Possible Integer overflow in OppositeSides.\n");
      printf("volume2 = %d     volume2d = %f\n",vol2,vol2d);
      exit(1);
    }
  
  if (((vol1 > 0) && (vol2 > 0)) || ((vol1 < 0) && (vol2 < 0)))
    return FALSE;
  
  else return TRUE;
}
/*
   The signed volumes of three tetrahedra are computed, determined
   by the segment ab, and each edge of the triangle.  As a kludge
   safety check, the volumes are computed both in integers and in
   doubles.
   Once the three volumes are in hand, whether there is an intersection,
   and if so the type, is determined by cases on the signs of the
   volumes.
   */
int DoesIntersect(Triangle T, pVertex a, pVertex b, boolean sharedsides)
{
  int vol1, vol2, vol3;
  double vol1d, vol2d, vol3d;
  
  vol1 = Volume6(a,T[0],T[1],b);
  vol1d= VolumeD(a,T[0],T[1],b);
  
  if (fabs (vol1 - vol1d) >= 1.0)
    {
      printf("Possible Integer overflow\n");
      exit(1);
    }
  vol2 = Volume6(a,T[1],T[2],b);
  vol2d= VolumeD(a,T[1],T[2],b);
  if (fabs (vol2 - vol2d) >= 1.0)
    {
      printf("Possible Integer overflow\n");
      exit(1);
    }
  vol3 = Volume6(a,T[2],T[0],b);
  vol3d= VolumeD(a,T[2],T[0],b);
  if (fabs (vol3 - vol3d) >= 1.0)
    {
      printf("Possible Integer overflow\n");
      exit(1);
    }
  /*
     printf("DoesIntersect:  Vol1 = %d   Vol2 = %d   Vol3 = %d\n",vol1,vol2,vol3);
     */
  
  /* intersects interior of triangle */
  if (((vol1 > 0) && (vol2 > 0) && (vol3 > 0)) || ((vol1 < 0) && 
						   (vol2 < 0) && (vol3 < 0)))
    return TRUE;
  
  /* no intersection */
  if (((vol1 > 0) || (vol2 > 0) || (vol3 > 0)) && ((vol1 < 0) || 
						   (vol2 < 0) || (vol3 < 0)))
    return FALSE;
  
  /* 3 zeros - coplanar */
  /* 
     NB:  It is quite possible for a segment coplanar with
     a triangle to intersect that triangle.  This happens, for
     example, when one triangle folds over back on top of a
     triangle adjacent to it.  So this case requires separate
     code.  Not yet implemented.  So these intersections are
     not detected!
     */
  else if ((vol1 == 0) && (vol2 == 0) && (vol3 == 0))
    {
      return coplanarCheck (T, a, b,sharedsides);
      
      
    }
  
  /* On vertex */
  else if (((vol1 == 0) && (vol2 == 0)) || ((vol1 == 0) && 
					    (vol3 == 0)) || ((vol2 == 0) && (vol3 == 0)))
    return FALSE;
  
  /* on edge */
  else if ((vol1 == 0) || (vol2 == 0) || (vol3 == 0))
    return TRUE;
  
  else
    return TRUE;
}

/*
   Returns volume of tetrahedron determined by f and p.
   Volume6 is positive iff p is on the negative side of f,
   where the positive side is determined by the rhrule.
   So volume is positive if ccw normal to f points outside
   the tetrahedron.
   [The expression could be factored to reduce the number
   of arithmetic operations.]
   */
int 	Volume6(pVertex a, pVertex b, pVertex c, pVertex p)
{
  int 	vol;
  int 	ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz;
  
  ax = a[0]; ay = a[1]; az = a[2]; 
  bx = b[0]; by = b[1]; bz = b[2]; 
  cx = c[0]; cy = c[1]; cz = c[2]; 
  dx = p[0]; dy = p[1]; dz = p[2];
  
  /*
     printf("ax = %d\tay=%d\taz=%d\n",ax,ay,az);
     printf("bx = %d\tby=%d\tbz=%d\n",bx,by,bz);
     printf("cx = %d\tcy=%d\tcz=%d\n",cx,cy,cz);
     printf("dx = %d\tdy=%d\tdz=%d\n",dx,dy,dz);
     */
  vol = -((az-dz)*(by-dy)*(cx-dx)) + (ay-dy)*(bz-dz)*(cx-dx) + 
    (az-dz)*(bx-dx)*(cy-dy) - (ax-dx)*(bz-dz)*(cy-dy) - 
      (ay-dy)*(bx-dx)*(cz-dz) + (ax-dx)*(by-dy)*(cz-dz);
	


  
  return vol;
}		

double VolumeD(pVertex a, pVertex b, pVertex c, pVertex p)
{
  double vol;
  double ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz;
  
  ax = a[0]; ay = a[1]; az = a[2]; 
  bx = b[0]; by = b[1]; bz = b[2]; 
  cx = c[0]; cy = c[1]; cz = c[2]; 
  dx = p[0]; dy = p[1]; dz = p[2];
  
  vol = -((az-dz)*(by-dy)*(cx-dx)) + (ay-dy)*(bz-dz)*(cx-dx) + 
    (az-dz)*(bx-dx)*(cy-dy) - (ax-dx)*(bz-dz)*(cy-dy) - 
      (ay-dy)*(bx-dx)*(cz-dz) + (ax-dx)*(by-dy)*(cz-dz);
	

  
  return vol;  
}		


void initIntersections (double intersections[3][3]) {

  int i,j;

  for (i = 0; i< 3; i++)
    for (j=0; j<3; j++)
      intersections[i][j] = 0;

}





/* coplanarCheck: this function checks if coplanar triangle and
   segment intersect. By Anna Lysyanskaya */

/* Method:

   1. Project the triangle and the segment to either xy, or yz, or zx plane:
       a. see if they already belong to one of the above planes, and
          choose that plane;
       b. otherwise, if the triangle is orthogonal to xy plane, change
          it to zx;
       c. define x- and y- coordinates in the plane we are using.
   2. In 2D, come up with the equation for the segment, and then 
      for each edge:
         find the equation (special case: x=const);
	 find where the edge line intersects with the segment line 
	     (special case: it they are the same or parallel).
   3. For each of the three intersections:
        if the intersection is inside of the corresponding edge
	   if the intersections is within the segment
              return TRUE.

   4. Determine if the segment is inside the triangle.
*/
	 
int coplanarCheck(Triangle T, pVertex a, pVertex b, boolean sharedsides) {
  
  pVertex vector1={0,0,1}, vector2, vector3, vector4;

  double y0, b0, y1, b1, intersections[3][3];

  int i, proj, x, y, j;

  initIntersections (intersections);

  /* Find out which projection to use */

  /* Can't project on z=0 if the triangle belongs to a plane orthogonal
        to z = 0. If it is x=const or y=const that the triangle belongs
	to, we have to use x=0 or y=0 respectively. Otherwise, the choice
	is arbitrary. */

  proj = 2;

  /* Check if this is x = const or y = const:  */
  for (i = 0; i < 3; i++) {

    if ((T[0][i] == T[1][i])&&(T[0][i] == T[2][i])) { 
      proj = i;
      goto compute;
    }
  }
  


  /* If a triangle is in a plane that is orthogonal to z=0, then a dot
     product of cross products of any two vectors on the planes 
     should be 0. vector1 = (0,0,1) */

  

  subtract (vector2, T[0], T[1]);
  subtract (vector3, T[0], T[2]);
  Cross(vector2, vector3, vector4);

  if (!Dot(vector1, vector4)) {

    /* We know that the plane is orthogonal to z = 0, and that it's not
       x = const or y = const. Then we can project everything to y=0. */
    proj = 1;

  }

  
  /* Find out where the intersections are: */
  
compute:  
  x = (proj +1)%3;  /* These are our not-really-x&y coordinates */
  y = (proj +2)%3;

  if (b[x]==a[x]) {

    /* the equation of this line is x=const. Swap x and y values: */
    i = x;
    x = y;
    y = i;

  }
  b0 = ((double) b[y]-a[y])/(b[x]-a[x]);
  y0 = a[y] - b0*a[x];
  
  /* Now we know the equation of the ab line: y = y0+b0*x */

  
  for (i = 0; i < 3; i++) {
    
    /* If the projection is selected correctly, and if the input
       is valid, there cannot be overflow. */
    if (((T[(i+1)%3][y]-T[i][y]) == 0)&&((T[(i+1)%3][x]-T[i][x])==0)) {
      printf ("Invalid input.\n");
      exit (1);

    }

    if ((T[(i+1)%3][x]-T[i][x])==0) {
      /* the equation of this line is x = const. */
      /* this time we can`t swap it that easily. */
      /* But we know that the intersection is when x=const=T[i][x] */
      intersections[i][0] = T[i][x];
      intersections[i][1] = y0 + b0 * T[i][x];
      /* printf ("a[x]=%d, b[x]=%d, T[%d][x]=%d, ") */
    }
    else {
      b1 = ((double) T[(i+1)%3][y]-T[i][y])/(T[(i+1)%3][x]-T[i][x]);
      y1 = T[i][y]-b1*T[i][x];
      if  (b0 != b1) {
	intersections[i][0] = (y1-y0)/(b0-b1);
	intersections[i][1] = y0 + b0*intersections[i][0];
	intersections[i][2] = 0;
      }
      else {    
	/* They are same or parallel: */
	if (y1 == y0) {
	  /* They are on the same line */
	  intersections[i][0] = T[i][x];
	  intersections[i][1] = T[i][y];
	  if ((!sharedsides) && inside(a,b,intersections[i], sharedsides,x,y))
	    return 1;

	  intersections[i][0] = T[(i+1)%3][x];
	  intersections[i][1] = T[(i+1)%3][y];
	  if ((!sharedsides) && inside(a,b,intersections[i], sharedsides,x,y))
	    return 1;
	}
	intersections[i][2] = 123;
      }
    }
  }
  
  /* Now, determine if the intersection occurs inside or outside: */
  for (j = 0; j < 3; j++) {

    if (inside(T[j], T[(j+1)%3], intersections[j], sharedsides, x, y)) {
        if (inside(a,b,intersections[j], sharedsides, x, y)) 
	  return 1;
      }
  }

  /* Now, determine if the diagonal is inside the triangle:  */


  if (surrounded (a,b, T, intersections, x, y))
    return 1;
  

  /* Since we haven't returned yet, there must be no intersection:  */

  return 0;



}



/* Surrounded: this function determines if a segment is inside a triangle
   when they are coplanar and we know the intersection points of the
   graphs of their functions. */
int surrounded (pVertex a, pVertex b, Triangle T, 
		double intersections[][3], int x, int y) {

  double vectors[2][2];
  int i;

  for (i = 0; i < 3; i ++) {
    
    if (inside (T[i], T[(i+1)%3], intersections[i], 0, x, y)) {
      if (inside (T[(i+1)%3], T[(i+2)%3], intersections[(i+1)%3], 0, x, y)) {
	if ((!inside (T[(i+2)%3], T[i], intersections[i], 0, x,y))&&
	    (!inside (T[(i+2)%3], T[i], intersections[(i+1)%3],0,x,y))) {
	  if (insideF (intersections[i], intersections[(i+1)%3], a, 0,
		       x,y)) {
	    if (insideF (intersections[i], intersections[(i+1)%3], b, 0,
			 x,y)){
	      return 1;
	    }
	  }
	}
      }
    }
  }
  return 0;

} 


/* inside: this function inputs endpoints of a segment and a point that 
   belongs to the same line, and determined whether the point is inside 
   of the segment. It also needs to know which are the presumed x- and
   y-coordinates. */

int inside (pVertex a, pVertex b, double intersection[3], int
	    sharedsides, int x, int y) {

  double vectors[2][2];
  int i;


  /* The third cell of this array is set to 123 iff the segment is parallel
     to the edge in question. That means, there's no intersection at all. */
  if (intersection[2] == 123) {
    return 0;
  }

  /* vector[0] = a-intersection;
     vector[1] = b-intersection; 
     if they point in different directions, then the intersections must
         be in between. */
  for (i = 0; i < 2; i++) {
    vectors[i][0] = ((i==0)?a[x]:b[x]) - intersection[0];
    vectors[i][1] = ((i==0)?a[y]:b[y]) - intersection[1];
    
    if ((vectors[i][0]==0)&&(vectors[i][1] ==0)) {
      if (sharedsides) return 0;
      else return 1;
      }
  }
  
  if (vectors[0][0] != 0)
    if (vectors[1][0]*vectors[0][0] > 0) return 0;
    else return 1;
  else
    if (vectors[1][1]*vectors[0][1] > 0) return 0;
    else return 1;

  

}



/* insideF: same as inside, but works for doubles. */
int insideF (double a[3], double b[3], pVertex intersection, int
	    sharedsides, int x, int y) {

  double vectors[2][2];
  int i;

  if ((a[2] == 123) || (b[2] == 123)) return 0;
     
  for (i = 0; i < 2; i++) {
    vectors[i][0] = ((i==0)?a[x]:b[x]) - intersection[0];
    vectors[i][1] = ((i==0)?a[y]:b[y]) - intersection[1];
    
    if ((vectors[i][0]==0)&&(vectors[i][1] ==0)) 
	if (sharedsides) return 0;
	else return 1;
  }

  if (vectors[0][0])
    if (vectors[1][0]/vectors[0][0] > 0) return 0;
    else return 1;
  else
    if (vectors[1][1]/vectors[0][1] > 0) return 0;
    else return 1;


}



/* subtract: vector X = a-b */
void subtract (pVertex X, pVertex a, pVertex b) {

  X[0] = a[0] - b[0];
  X[1] = a[1] - b[1];
  X[2] = a[2] - b[2];


}



/*------------------------------------------------------------------*/
/*                         Vector Functions                         */
/*------------------------------------------------------------------*/
/*
	Deposits a cross b into c.
*/
void	Cross( pVertex a, pVertex b, pVertex c)
{
	int	i;

	for (i = 0; i<3; i++ )
		c[i] =
			a[(i+1) % 3] * b[(i+2) % 3] -
			a[(i+2) % 3] * b[(i+1) % 3];
}
/*------------------------------------------------------------------*/
/*
	Computes the dot product.
*/
int Dot( pVertex a, pVertex b )
{
	int	i;
	int	sum;

	sum = 0;

	for( i=0; i<3; i++ )
		sum = sum + a[i] * b[i];
	return sum;
}
/*------------------------------------------------------------------*/
